@@ -1,17 +1,13 @@
-use bitcoin::{BlockHash, Transaction, Txid};
+use bitcoin::{Transaction, Txid};
 use parking_lot::RwLock;
 
 use std::collections::HashMap;
 use std::sync::Arc;
 
-use crate::{
-    merkle::Proof,
-    metrics::{self, Histogram, Metrics},
-};
+use crate::metrics::{self, Histogram, Metrics};
 
 pub(crate) struct Cache {
     txs: Arc<RwLock<HashMap<Txid, Transaction>>>,
-    proofs: Arc<RwLock<HashMap<(BlockHash, Txid), Proof>>>,
 
     // stats
     txs_size: Histogram,
@@ -21,7 +17,6 @@ impl Cache {
     pub fn new(metrics: &Metrics) -> Self {
         Cache {
             txs: Default::default(),
-            proofs: Default::default(),
             txs_size: metrics.histogram_vec(
                 "cache_txs_size",
                 "Cached transactions' size (in bytes)",
@@ -45,21 +40,4 @@ impl Cache {
     {
         self.txs.read().get(txid).map(f)
     }
-
-    pub fn add_proof<F>(&self, blockhash: BlockHash, txid: Txid, f: F)
-    where
-        F: FnOnce() -> Proof,
-    {
-        self.proofs
-            .write()
-            .entry((blockhash, txid))
-            .or_insert_with(f);
-    }
-
-    pub fn get_proof<F, T>(&self, blockhash: BlockHash, txid: Txid, f: F) -> Option<T>
-    where
-        F: FnOnce(&Proof) -> T,
-    {
-        self.proofs.read().get(&(blockhash, txid)).map(f)
-    }
 }
