@@ -17,8 +17,9 @@ const DEFAULT_SERVER_ADDRESS: [u8; 4] = [127, 0, 0, 1]; // by default, serve on
 
 mod internal {
     #![allow(unused)]
-    #![allow(clippy::identity_conversion)]
     #![allow(clippy::cognitive_complexity)]
+    #![allow(clippy::unnecessary_lazy_evaluations)]
+    #![allow(clippy::useless_conversion)]
 
     include!(concat!(env!("OUT_DIR"), "/configure_me_config.rs"));
 }
@@ -176,23 +177,32 @@ fn default_daemon_dir() -> PathBuf {
     home
 }
 
+fn default_config_files() -> Vec<OsString> {
+    #[cfg(not(feature = "ignore_default_config_files"))]
+    {
+        let mut files = vec![OsString::from("electrs.toml")]; // cwd
+        if let Some(mut path) = home_dir() {
+            path.extend(&[".electrs", "config.toml"]);
+            files.push(OsString::from(path)) // home directory
+        }
+        files.push(OsString::from("/etc/electrs/config.toml")); // system-wide
+        files
+    }
+
+    #[cfg(feature = "ignore_default_config_files")]
+    {
+        vec![]
+    }
+}
+
 impl Config {
     /// Parses args, env vars, config files and post-processes them
     pub fn from_args() -> Config {
         use internal::ResultExt;
 
-        let system_config: &OsStr = "/etc/electrs/config.toml".as_ref();
-        let home_config = home_dir().map(|mut dir| {
-            dir.extend(&[".electrs", "config.toml"]);
-            dir
-        });
-        let cwd_config: &OsStr = "electrs.toml".as_ref();
-        let configs = std::iter::once(cwd_config)
-            .chain(home_config.as_ref().map(AsRef::as_ref))
-            .chain(std::iter::once(system_config));
-
         let (mut config, args) =
-            internal::Config::including_optional_config_files(configs).unwrap_or_exit();
+            internal::Config::including_optional_config_files(default_config_files())
+                .unwrap_or_exit();
 
         let db_subdir = match config.network {
             Network::Bitcoin => "bitcoin",
@@ -277,12 +287,11 @@ impl Config {
             }
         });
 
-        let level = match config.verbose {
-            0 => log::LevelFilter::Error,
-            1 => log::LevelFilter::Warn,
-            2 => log::LevelFilter::Info,
-            _ => log::LevelFilter::Debug,
-        };
+        if config.verbose > 0 {
+            eprintln!("Error: please use `log_filters` to set logging verbosity",);
+            std::process::exit(1);
+        }
+        let log_filters = config.log_filters;
 
         let index_lookup_limit = match config.index_lookup_limit {
             0 => None,
@@ -297,6 +306,11 @@ impl Config {
             std::process::exit(1);
         }
 
+        if config.version {
+            println!("v{}", ELECTRS_VERSION);
+            std::process::exit(0);
+        }
+
         let config = Config {
             network: config.network,
             db_path: config.db_dir,
@@ -322,11 +336,13 @@ impl Config {
             "Starting electrs {} on {} {} with {:?}",
             ELECTRS_VERSION, ARCH, OS, config
         );
-        env_logger::Builder::from_default_env()
-            .default_format()
-            .format_timestamp_millis()
-            .filter_level(level)
-            .init();
+        let mut builder = env_logger::Builder::from_default_env();
+        builder.default_format().format_timestamp_millis();
+        if let Some(log_filters) = &log_filters {
+            builder.parse_filters(log_filters);
+        }
+        builder.init();
+
         config
     }
 }
