@@ -17,7 +17,7 @@ use crate::{
     config::{Config, ELECTRS_VERSION},
     daemon::{self, extract_bitcoind_error, Daemon},
     merkle::Proof,
-    metrics::{self, Histogram},
+    metrics::{self, Histogram, Metrics},
     signals::Signal,
     status::ScriptHashStatus,
     tracker::Tracker,
@@ -27,6 +27,8 @@ use crate::{
 const PROTOCOL_VERSION: &str = "1.4";
 const UNKNOWN_FEE: isize = -1; // (allowed by Electrum protocol)
 
+const UNSUBSCRIBED_QUERY_MESSAGE: &str = "your wallet uses less efficient method of querying electrs, consider contacting the developer of your wallet. Reason:";
+
 /// Per-client Electrum protocol state
 #[derive(Default)]
 pub struct Client {
@@ -64,11 +66,11 @@ enum TxGetArgs {
     TxidVerbose(Txid, bool),
 }
 
-impl From<TxGetArgs> for (Txid, bool) {
-    fn from(args: TxGetArgs) -> Self {
+impl From<&TxGetArgs> for (Txid, bool) {
+    fn from(args: &TxGetArgs) -> Self {
         match args {
-            TxGetArgs::Txid((txid,)) => (txid, false),
-            TxGetArgs::TxidVerbose(txid, verbose) => (txid, verbose),
+            TxGetArgs::Txid((txid,)) => (*txid, false),
+            TxGetArgs::TxidVerbose(txid, verbose) => (*txid, *verbose),
         }
     }
 }
@@ -86,6 +88,7 @@ enum RpcError {
     // Electrum-specific errors
     BadRequest(anyhow::Error),
     DaemonError(daemon::RpcError),
+    UnavailableIndex,
 }
 
 impl RpcError {
@@ -105,6 +108,10 @@ impl RpcError {
             },
             RpcError::BadRequest(err) => json!({"code": 1, "message": err.to_string()}),
             RpcError::DaemonError(err) => json!({"code": 2, "message": err.message}),
+            RpcError::UnavailableIndex => {
+                // Internal JSON-RPC error (https://www.jsonrpc.org/specification#error_object)
+                json!({"code": -32603, "message": "unavailable index"})
+            }
         }
     }
 }
@@ -122,15 +129,15 @@ pub struct Rpc {
 
 impl Rpc {
     /// Perform initial index sync (may take a while on first run).
-    pub fn new(config: &Config) -> Result<Self> {
-        let tracker = Tracker::new(&config)?;
-        let rpc_duration = tracker.metrics().histogram_vec(
+    pub fn new(config: &Config, metrics: Metrics) -> Result<Self> {
+        let rpc_duration = metrics.histogram_vec(
             "rpc_duration",
             "RPC duration (in seconds)",
             "method",
             metrics::default_duration_buckets(),
         );
 
+        let tracker = Tracker::new(config, metrics)?;
         let signal = Signal::new();
         let daemon = Daemon::connect(config, signal.exit_flag(), tracker.metrics())?;
         let cache = Cache::new(tracker.metrics());
@@ -153,7 +160,7 @@ impl Rpc {
         self.daemon.new_block_notification()
     }
 
-    pub fn sync(&mut self) -> Result<()> {
+    pub fn sync(&mut self) -> Result<bool> {
         self.tracker.sync(&self.daemon, self.signal.exit_flag())
     }
 
@@ -241,16 +248,16 @@ impl Rpc {
     fn scripthash_get_balance(
         &self,
         client: &Client,
-        (scripthash,): (ScriptHash,),
+        (scripthash,): &(ScriptHash,),
     ) -> Result<Value> {
-        let balance = match client.scripthashes.get(&scripthash) {
+        let balance = match client.scripthashes.get(scripthash) {
             Some(status) => self.tracker.get_balance(status),
             None => {
-                warn!(
-                    "blockchain.scripthash.get_balance called for unsubscribed scripthash: {}",
-                    scripthash
+                info!(
+                    "{} blockchain.scripthash.get_balance called for unsubscribed scripthash: {}",
+                    UNSUBSCRIBED_QUERY_MESSAGE, scripthash
                 );
-                self.tracker.get_balance(&self.new_status(scripthash)?)
+                self.tracker.get_balance(&self.new_status(*scripthash)?)
             }
         };
         Ok(json!(balance))
@@ -259,16 +266,16 @@ impl Rpc {
     fn scripthash_get_history(
         &self,
         client: &Client,
-        (scripthash,): (ScriptHash,),
+        (scripthash,): &(ScriptHash,),
     ) -> Result<Value> {
-        let history_entries = match client.scripthashes.get(&scripthash) {
+        let history_entries = match client.scripthashes.get(scripthash) {
             Some(status) => json!(status.get_history()),
             None => {
-                warn!(
-                    "blockchain.scripthash.get_history called for unsubscribed scripthash: {}",
-                    scripthash
+                info!(
+                    "{} blockchain.scripthash.get_history called for unsubscribed scripthash: {}",
+                    UNSUBSCRIBED_QUERY_MESSAGE, scripthash
                 );
-                json!(self.new_status(scripthash)?.get_history())
+                json!(self.new_status(*scripthash)?.get_history())
             }
         };
         Ok(history_entries)
@@ -277,16 +284,16 @@ impl Rpc {
     fn scripthash_list_unspent(
         &self,
         client: &Client,
-        (scripthash,): (ScriptHash,),
+        (scripthash,): &(ScriptHash,),
     ) -> Result<Value> {
-        let unspent_entries = match client.scripthashes.get(&scripthash) {
+        let unspent_entries = match client.scripthashes.get(scripthash) {
             Some(status) => self.tracker.get_unspent(status),
             None => {
-                warn!(
-                    "blockchain.scripthash.listunspent called for unsubscribed scripthash: {}",
-                    scripthash
+                info!(
+                    "{} blockchain.scripthash.listunspent called for unsubscribed scripthash: {}",
+                    UNSUBSCRIBED_QUERY_MESSAGE, scripthash
                 );
-                self.tracker.get_unspent(&self.new_status(scripthash)?)
+                self.tracker.get_unspent(&self.new_status(*scripthash)?)
             }
         };
         Ok(json!(unspent_entries))
@@ -295,13 +302,41 @@ impl Rpc {
     fn scripthash_subscribe(
         &self,
         client: &mut Client,
-        (scripthash,): (ScriptHash,),
+        (scripthash,): &(ScriptHash,),
     ) -> Result<Value> {
-        let result = match client.scripthashes.entry(scripthash) {
-            Entry::Occupied(e) => e.get().statushash(),
-            Entry::Vacant(e) => e.insert(self.new_status(scripthash)?).statushash(),
-        };
-        Ok(json!(result))
+        self.scripthashes_subscribe(client, &[*scripthash])
+            .next()
+            .unwrap()
+    }
+
+    fn scripthashes_subscribe<'a>(
+        &self,
+        client: &'a mut Client,
+        scripthashes: &'a [ScriptHash],
+    ) -> impl Iterator<Item = Result<Value>> + 'a {
+        let new_scripthashes: Vec<ScriptHash> = scripthashes
+            .iter()
+            .copied()
+            .filter(|scripthash| !client.scripthashes.contains_key(scripthash))
+            .collect();
+
+        let mut results: HashMap<ScriptHash, Result<ScriptHashStatus>> = new_scripthashes
+            .into_par_iter()
+            .map(|scripthash| (scripthash, self.new_status(scripthash)))
+            .collect();
+
+        scripthashes.iter().map(move |scripthash| {
+            let statushash = match client.scripthashes.entry(*scripthash) {
+                Entry::Occupied(e) => e.get().statushash(),
+                Entry::Vacant(e) => {
+                    let status = results
+                        .remove(scripthash)
+                        .expect("missing scripthash status")?; // return an error for failed subscriptions
+                    e.insert(status).statushash()
+                }
+            };
+            Ok(json!(statushash))
+        })
     }
 
     fn new_status(&self, scripthash: ScriptHash) -> Result<ScriptHashStatus> {
@@ -311,14 +346,14 @@ impl Rpc {
         Ok(status)
     }
 
-    fn transaction_broadcast(&self, (tx_hex,): (String,)) -> Result<Value> {
+    fn transaction_broadcast(&self, (tx_hex,): &(String,)) -> Result<Value> {
         let tx_bytes = Vec::from_hex(&tx_hex).context("non-hex transaction")?;
         let tx = deserialize(&tx_bytes).context("invalid transaction")?;
         let txid = self.daemon.broadcast(&tx)?;
         Ok(json!(txid))
     }
 
-    fn transaction_get(&self, args: TxGetArgs) -> Result<Value> {
+    fn transaction_get(&self, args: &TxGetArgs) -> Result<Value> {
         let (txid, verbose) = args.into();
         if verbose {
             let blockhash = self.tracker.get_blockhash_by_txid(txid);
@@ -328,34 +363,30 @@ impl Rpc {
         Ok(match cached {
             Some(tx_hex) => json!(tx_hex),
             None => {
-                debug!("tx cache miss: {}", txid);
+                debug!("tx cache miss: txid={}", txid);
                 let blockhash = self.tracker.get_blockhash_by_txid(txid);
                 json!(self.daemon.get_transaction_hex(&txid, blockhash)?)
             }
         })
     }
 
-    fn transaction_get_merkle(&self, (txid, height): (Txid, usize)) -> Result<Value> {
+    fn transaction_get_merkle(&self, (txid, height): &(Txid, usize)) -> Result<Value> {
         let chain = self.tracker.chain();
-        let blockhash = match chain.get_block_hash(height) {
+        let blockhash = match chain.get_block_hash(*height) {
             None => bail!("missing block at {}", height),
             Some(blockhash) => blockhash,
         };
-        let proof_to_value = |proof: &Proof| {
-            json!({
+        let txids = self.daemon.get_block_txids(blockhash)?;
+        match txids.iter().position(|current_txid| *current_txid == *txid) {
+            None => bail!("missing txid {} in block {}", txid, blockhash),
+            Some(position) => {
+                let proof = Proof::create(&txids, position);
+                Ok(json!({
                 "block_height": height,
                 "pos": proof.position(),
                 "merkle": proof.to_hex(),
-            })
-        };
-        if let Some(result) = self.cache.get_proof(blockhash, txid, proof_to_value) {
-            return Ok(result);
-        }
-        debug!("txids cache miss: {}", blockhash);
-        let txids = self.daemon.get_block_txids(blockhash)?;
-        match txids.iter().position(|current_txid| *current_txid == txid) {
-            None => bail!("missing txid {} in block {}", txid, blockhash),
-            Some(position) => Ok(proof_to_value(&Proof::create(&txids, position))),
+                }))
+            }
         }
     }
 
@@ -367,7 +398,7 @@ impl Rpc {
         format!("electrs/{}", ELECTRS_VERSION)
     }
 
-    fn version(&self, (client_id, client_version): (String, Version)) -> Result<Value> {
+    fn version(&self, (client_id, client_version): &(String, Version)) -> Result<Value> {
         match client_version {
             Version::Single(v) if v == PROTOCOL_VERSION => {
                 Ok(json!([self.server_id(), PROTOCOL_VERSION]))
@@ -395,79 +426,117 @@ impl Rpc {
         }))
     }
 
-    pub fn handle_request(&self, client: &mut Client, line: &str) -> String {
-        let error_msg_no_id = |err| error_msg(Value::Null, RpcError::Standard(err));
-        let response: Value = match serde_json::from_str(line) {
-            // parse JSON from str
-            Ok(value) => match serde_json::from_value(value) {
-                // parse RPC from JSON
-                Ok(requests) => match requests {
-                    Requests::Single(request) => self.call(client, request),
-                    Requests::Batch(requests) => json!(requests
-                        .into_iter()
-                        .map(|request| self.call(client, request))
-                        .collect::<Vec<Value>>()),
-                },
-                Err(err) => {
-                    warn!("invalid RPC request ({:?}): {}", line, err);
-                    error_msg_no_id(StandardError::InvalidRequest)
-                }
-            },
-            Err(err) => {
-                warn!("invalid JSON ({:?}): {}", line, err);
-                error_msg_no_id(StandardError::ParseError)
-            }
-        };
-        response.to_string()
+    pub fn handle_requests(&self, client: &mut Client, lines: &[String]) -> Vec<String> {
+        let parsed: Vec<Result<Calls, Value>> = lines
+            .iter()
+            .map(|line| {
+                parse_requests(&line)
+                    .map(Calls::parse)
+                    .map_err(error_msg_no_id)
+            })
+            .collect();
+
+        parsed
+            .into_iter()
+            .map(|calls| self.handle_calls(client, calls).to_string())
+            .collect()
     }
 
-    fn call(&self, client: &mut Client, request: Request) -> Value {
-        let Request { id, method, params } = request;
-        let call = match Call::parse(&method, params) {
-            Ok(call) => call,
-            Err(err) => return error_msg(id, RpcError::Standard(err)),
+    fn handle_calls(&self, client: &mut Client, calls: Result<Calls, Value>) -> Value {
+        let calls: Calls = match calls {
+            Ok(calls) => calls,
+            Err(response) => return response, // JSON parsing failed - the response does not contain request id
         };
-        self.rpc_duration.observe_duration(&method, || {
-            let result = match call {
-                Call::Banner => Ok(json!(self.banner)),
-                Call::BlockHeader(args) => self.block_header(args),
-                Call::BlockHeaders(args) => self.block_headers(args),
-                Call::Donation => Ok(Value::Null),
-                Call::EstimateFee(args) => self.estimate_fee(args),
-                Call::Features => self.features(),
-                Call::HeadersSubscribe => self.headers_subscribe(client),
-                Call::MempoolFeeHistogram => self.get_fee_histogram(),
-                Call::PeersSubscribe => Ok(json!([])),
-                Call::Ping => Ok(Value::Null),
-                Call::RelayFee => self.relayfee(),
-                Call::ScriptHashGetBalance(args) => self.scripthash_get_balance(client, args),
-                Call::ScriptHashGetHistory(args) => self.scripthash_get_history(client, args),
-                Call::ScriptHashListUnspent(args) => self.scripthash_list_unspent(client, args),
-                Call::ScriptHashSubscribe(args) => self.scripthash_subscribe(client, args),
-                Call::TransactionBroadcast(args) => self.transaction_broadcast(args),
-                Call::TransactionGet(args) => self.transaction_get(args),
-                Call::TransactionGetMerkle(args) => self.transaction_get_merkle(args),
-                Call::Version(args) => self.version(args),
-            };
-            match result {
-                Ok(value) => result_msg(id, value),
-                Err(err) => {
-                    warn!("RPC {} failed: {:#}", method, err);
-                    match err
-                        .downcast_ref::<bitcoincore_rpc::Error>()
-                        .and_then(extract_bitcoind_error)
-                    {
-                        Some(e) => error_msg(id, RpcError::DaemonError(e.clone())),
-                        None => error_msg(id, RpcError::BadRequest(err)),
-                    }
+
+        match calls {
+            Calls::Batch(batch) => {
+                if let Some(result) = self.try_multi_call(client, &batch) {
+                    return json!(result);
                 }
+                json!(batch
+                    .into_iter()
+                    .map(|result| self.single_call(client, result))
+                    .collect::<Vec<Value>>())
             }
+            Calls::Single(result) => self.single_call(client, result),
+        }
+    }
+
+    fn try_multi_call(
+        &self,
+        client: &mut Client,
+        calls: &[Result<Call, Value>],
+    ) -> Option<Vec<Value>> {
+        // exit if any call failed to parse
+        let valid_calls = calls
+            .iter()
+            .map(|result| result.as_ref().ok())
+            .collect::<Option<Vec<&Call>>>()?;
+
+        // only "blockchain.scripthashes.subscribe" are supported
+        let scripthashes: Vec<ScriptHash> = valid_calls
+            .iter()
+            .map(|call| match &call.params {
+                Params::ScriptHashSubscribe((scripthash,)) => Some(*scripthash),
+                _ => None, // exit if any of the calls is not supported
+            })
+            .collect::<Option<Vec<ScriptHash>>>()?;
+
+        Some(
+            self.rpc_duration
+                .observe_duration("blockchain.scripthash.subscribe:multi", || {
+                    self.scripthashes_subscribe(client, &scripthashes)
+                        .zip(valid_calls)
+                        .map(|(result, call)| call.response(result))
+                        .collect::<Vec<Value>>()
+                }),
+        )
+    }
+
+    fn single_call(&self, client: &mut Client, call: Result<Call, Value>) -> Value {
+        let call = match call {
+            Ok(call) => call,
+            Err(response) => return response, // params parsing may fail - the response contains request id
+        };
+        self.rpc_duration.observe_duration(&call.method, || {
+            if self.tracker.status().is_err() {
+                // Allow only a few RPC (for sync status notification) not requiring index DB being compacted.
+                match &call.params {
+                    Params::BlockHeader(_)
+                    | Params::BlockHeaders(_)
+                    | Params::HeadersSubscribe
+                    | Params::Version(_) => (),
+                    _ => return error_msg(&call.id, RpcError::UnavailableIndex),
+                };
+            }
+            let result = match &call.params {
+                Params::Banner => Ok(json!(self.banner)),
+                Params::BlockHeader(args) => self.block_header(*args),
+                Params::BlockHeaders(args) => self.block_headers(*args),
+                Params::Donation => Ok(Value::Null),
+                Params::EstimateFee(args) => self.estimate_fee(*args),
+                Params::Features => self.features(),
+                Params::HeadersSubscribe => self.headers_subscribe(client),
+                Params::MempoolFeeHistogram => self.get_fee_histogram(),
+                Params::PeersSubscribe => Ok(json!([])),
+                Params::Ping => Ok(Value::Null),
+                Params::RelayFee => self.relayfee(),
+                Params::ScriptHashGetBalance(args) => self.scripthash_get_balance(client, args),
+                Params::ScriptHashGetHistory(args) => self.scripthash_get_history(client, args),
+                Params::ScriptHashListUnspent(args) => self.scripthash_list_unspent(client, args),
+                Params::ScriptHashSubscribe(args) => self.scripthash_subscribe(client, args),
+                Params::TransactionBroadcast(args) => self.transaction_broadcast(args),
+                Params::TransactionGet(args) => self.transaction_get(args),
+                Params::TransactionGetMerkle(args) => self.transaction_get_merkle(args),
+                Params::Version(args) => self.version(args),
+            };
+            call.response(result)
         })
     }
 }
 
 #[derive(Deserialize)]
-enum Call {
+enum Params {
     Banner,
     BlockHeader((usize,)),
     BlockHeaders((usize, usize)),
@@ -489,28 +558,28 @@ enum Call {
     Version((String, Version)),
 }
 
-impl Call {
-    fn parse(method: &str, params: Value) -> std::result::Result<Call, StandardError> {
+impl Params {
+    fn parse(method: &str, params: Value) -> std::result::Result<Params, StandardError> {
         Ok(match method {
-            "blockchain.block.header" => Call::BlockHeader(convert(params)?),
-            "blockchain.block.headers" => Call::BlockHeaders(convert(params)?),
-            "blockchain.estimatefee" => Call::EstimateFee(convert(params)?),
-            "blockchain.headers.subscribe" => Call::HeadersSubscribe,
-            "blockchain.relayfee" => Call::RelayFee,
-            "blockchain.scripthash.get_balance" => Call::ScriptHashGetBalance(convert(params)?),
-            "blockchain.scripthash.get_history" => Call::ScriptHashGetHistory(convert(params)?),
-            "blockchain.scripthash.listunspent" => Call::ScriptHashListUnspent(convert(params)?),
-            "blockchain.scripthash.subscribe" => Call::ScriptHashSubscribe(convert(params)?),
-            "blockchain.transaction.broadcast" => Call::TransactionBroadcast(convert(params)?),
-            "blockchain.transaction.get" => Call::TransactionGet(convert(params)?),
-            "blockchain.transaction.get_merkle" => Call::TransactionGetMerkle(convert(params)?),
-            "mempool.get_fee_histogram" => Call::MempoolFeeHistogram,
-            "server.banner" => Call::Banner,
-            "server.donation_address" => Call::Donation,
-            "server.features" => Call::Features,
-            "server.peers.subscribe" => Call::PeersSubscribe,
-            "server.ping" => Call::Ping,
-            "server.version" => Call::Version(convert(params)?),
+            "blockchain.block.header" => Params::BlockHeader(convert(params)?),
+            "blockchain.block.headers" => Params::BlockHeaders(convert(params)?),
+            "blockchain.estimatefee" => Params::EstimateFee(convert(params)?),
+            "blockchain.headers.subscribe" => Params::HeadersSubscribe,
+            "blockchain.relayfee" => Params::RelayFee,
+            "blockchain.scripthash.get_balance" => Params::ScriptHashGetBalance(convert(params)?),
+            "blockchain.scripthash.get_history" => Params::ScriptHashGetHistory(convert(params)?),
+            "blockchain.scripthash.listunspent" => Params::ScriptHashListUnspent(convert(params)?),
+            "blockchain.scripthash.subscribe" => Params::ScriptHashSubscribe(convert(params)?),
+            "blockchain.transaction.broadcast" => Params::TransactionBroadcast(convert(params)?),
+            "blockchain.transaction.get" => Params::TransactionGet(convert(params)?),
+            "blockchain.transaction.get_merkle" => Params::TransactionGetMerkle(convert(params)?),
+            "mempool.get_fee_histogram" => Params::MempoolFeeHistogram,
+            "server.banner" => Params::Banner,
+            "server.donation_address" => Params::Donation,
+            "server.features" => Params::Features,
+            "server.peers.subscribe" => Params::PeersSubscribe,
+            "server.ping" => Params::Ping,
+            "server.version" => Params::Version(convert(params)?),
             _ => {
                 warn!("unknown method {}", method);
                 return Err(StandardError::MethodNotFound);
@@ -519,6 +588,57 @@ impl Call {
     }
 }
 
+struct Call {
+    id: Value,
+    method: String,
+    params: Params,
+}
+
+impl Call {
+    fn parse(request: Request) -> Result<Call, Value> {
+        match Params::parse(&request.method, request.params) {
+            Ok(params) => Ok(Call {
+                id: request.id,
+                method: request.method,
+                params,
+            }),
+            Err(e) => Err(error_msg(&request.id, RpcError::Standard(e))),
+        }
+    }
+
+    fn response(&self, result: Result<Value>) -> Value {
+        match result {
+            Ok(value) => result_msg(&self.id, value),
+            Err(err) => {
+                warn!("RPC {} failed: {:#}", self.method, err);
+                match err
+                    .downcast_ref::<bitcoincore_rpc::Error>()
+                    .and_then(extract_bitcoind_error)
+                {
+                    Some(e) => error_msg(&self.id, RpcError::DaemonError(e.clone())),
+                    None => error_msg(&self.id, RpcError::BadRequest(err)),
+                }
+            }
+        }
+    }
+}
+
+enum Calls {
+    Batch(Vec<Result<Call, Value>>),
+    Single(Result<Call, Value>),
+}
+
+impl Calls {
+    fn parse(requests: Requests) -> Calls {
+        match requests {
+            Requests::Single(request) => Calls::Single(Call::parse(request)),
+            Requests::Batch(batch) => {
+                Calls::Batch(batch.into_iter().map(Call::parse).collect::<Vec<_>>())
+            }
+        }
+    }
+}
+
 fn convert<T>(params: Value) -> std::result::Result<T, StandardError>
 where
     T: serde::de::DeserializeOwned,
@@ -534,10 +654,32 @@ fn notification(method: &str, params: &[Value]) -> Value {
     json!({"jsonrpc": "2.0", "method": method, "params": params})
 }
 
-fn result_msg(id: Value, result: Value) -> Value {
+fn result_msg(id: &Value, result: Value) -> Value {
     json!({"jsonrpc": "2.0", "id": id, "result": result})
 }
 
-fn error_msg(id: Value, error: RpcError) -> Value {
+fn error_msg(id: &Value, error: RpcError) -> Value {
     json!({"jsonrpc": "2.0", "id": id, "error": error.to_value()})
 }
+
+fn error_msg_no_id(err: StandardError) -> Value {
+    error_msg(&Value::Null, RpcError::Standard(err))
+}
+
+fn parse_requests(line: &str) -> Result<Requests, StandardError> {
+    match serde_json::from_str(line) {
+        // parse JSON from str
+        Ok(value) => match serde_json::from_value(value) {
+            // parse RPC from JSON
+            Ok(requests) => Ok(requests),
+            Err(err) => {
+                warn!("invalid RPC request ({:?}): {}", line, err);
+                Err(StandardError::InvalidRequest)
+            }
+        },
+        Err(err) => {
+            warn!("invalid JSON ({:?}): {}", line, err);
+            Err(StandardError::ParseError)
+        }
+    }
+}
