@@ -3,12 +3,12 @@ use bitcoin::consensus::{deserialize, serialize};
 use bitcoin::{Block, BlockHash, OutPoint, Txid};
 
 use crate::{
-    chain::Chain,
+    chain::{Chain, NewHeader},
     daemon::Daemon,
     db::{DBStore, Row, WriteBatch},
     metrics::{self, Gauge, Histogram, Metrics},
     signals::ExitFlag,
-    types::{HeaderRow, ScriptHash, ScriptHashRow, SpendingPrefixRow, TxidRow},
+    types::{HashPrefixRow, HeaderRow, ScriptHash, ScriptHashRow, SpendingPrefixRow, TxidRow},
 };
 
 #[derive(Clone)]
@@ -16,7 +16,7 @@ struct Stats {
     update_duration: Histogram,
     update_size: Histogram,
     height: Gauge,
-    db_size: Gauge,
+    db_properties: Gauge,
 }
 
 impl Stats {
@@ -35,27 +35,60 @@ impl Stats {
                 metrics::default_size_buckets(),
             ),
             height: metrics.gauge("index_height", "Indexed block height", "type"),
-            db_size: metrics.gauge("index_db_size", "Index DB size (bytes)", "type"),
+            db_properties: metrics.gauge("index_db_properties", "Index DB properties", "name"),
+        }
+    }
+
+    fn observe_duration<T>(&self, label: &str, f: impl FnOnce() -> T) -> T {
+        self.update_duration.observe_duration(label, f)
+    }
+
+    fn observe_size(&self, label: &str, rows: &[Row]) {
+        self.update_size.observe(label, db_rows_size(rows));
+    }
+
+    fn observe_batch(&self, batch: &WriteBatch) {
+        self.observe_size("write_funding_rows", &batch.funding_rows);
+        self.observe_size("write_spending_rows", &batch.spending_rows);
+        self.observe_size("write_txid_rows", &batch.txid_rows);
+        self.observe_size("write_header_rows", &batch.header_rows);
+        debug!(
+            "writing {} funding and {} spending rows from {} transactions, {} blocks",
+            batch.funding_rows.len(),
+            batch.spending_rows.len(),
+            batch.txid_rows.len(),
+            batch.header_rows.len()
+        );
+    }
+
+    fn observe_chain(&self, chain: &Chain) {
+        self.height.set("tip", chain.height() as f64);
+    }
+
+    fn observe_db(&self, store: &DBStore) {
+        for (cf, name, value) in store.get_properties() {
+            self.db_properties
+                .set(&format!("{}:{}", name, cf), value as f64);
         }
     }
 }
 
 struct IndexResult {
     header_row: HeaderRow,
-    funding_rows: Vec<ScriptHashRow>,
-    spending_rows: Vec<SpendingPrefixRow>,
-    txid_rows: Vec<TxidRow>,
+    funding_rows: Vec<HashPrefixRow>,
+    spending_rows: Vec<HashPrefixRow>,
+    txid_rows: Vec<HashPrefixRow>,
 }
 
 impl IndexResult {
     fn extend(&self, batch: &mut WriteBatch) {
-        let funding_rows = self.funding_rows.iter().map(ScriptHashRow::to_db_row);
+        let funding_rows = self.funding_rows.iter().map(HashPrefixRow::to_db_row);
         batch.funding_rows.extend(funding_rows);
 
-        let spending_rows = self.spending_rows.iter().map(SpendingPrefixRow::to_db_row);
+        let spending_rows = self.spending_rows.iter().map(HashPrefixRow::to_db_row);
         batch.spending_rows.extend(spending_rows);
 
-        let txid_rows = self.txid_rows.iter().map(TxidRow::to_db_row);
+        let txid_rows = self.txid_rows.iter().map(HashPrefixRow::to_db_row);
         batch.txid_rows.extend(txid_rows);
 
         batch.header_rows.push(self.header_row.to_db_row());
@@ -70,6 +103,7 @@ pub struct Index {
     lookup_limit: Option<usize>,
     chain: Chain,
     stats: Stats,
+    is_ready: bool,
 }
 
 impl Index {
@@ -91,17 +125,16 @@ impl Index {
             chain.load(headers, tip);
             chain.drop_last_headers(reindex_last_blocks);
         };
-
         let stats = Stats::new(metrics);
-        stats.height.set("tip", chain.height() as f64);
-        stats.db_size.set("total", store.get_size()? as f64);
-
+        stats.observe_chain(&chain);
+        stats.observe_db(&store);
         Ok(Index {
             store,
             batch_size,
             lookup_limit,
             chain,
             stats,
+            is_ready: false,
         })
     }
 
@@ -124,7 +157,7 @@ impl Index {
     pub(crate) fn filter_by_txid(&self, txid: Txid) -> impl Iterator<Item = BlockHash> + '_ {
         self.store
             .iter_txid(TxidRow::scan_prefix(txid))
-            .map(|row| TxidRow::from_db_row(&row).height())
+            .map(|row| HashPrefixRow::from_db_row(&row).height())
             .filter_map(move |height| self.chain.get_block_hash(height))
     }
 
@@ -134,7 +167,7 @@ impl Index {
     ) -> impl Iterator<Item = BlockHash> + '_ {
         self.store
             .iter_funding(ScriptHashRow::scan_prefix(scripthash))
-            .map(|row| ScriptHashRow::from_db_row(&row).height())
+            .map(|row| HashPrefixRow::from_db_row(&row).height())
             .filter_map(move |height| self.chain.get_block_hash(height))
     }
 
@@ -144,84 +177,74 @@ impl Index {
     ) -> impl Iterator<Item = BlockHash> + '_ {
         self.store
             .iter_spending(SpendingPrefixRow::scan_prefix(outpoint))
-            .map(|row| SpendingPrefixRow::from_db_row(&row).height())
+            .map(|row| HashPrefixRow::from_db_row(&row).height())
             .filter_map(move |height| self.chain.get_block_hash(height))
     }
 
-    fn observe_duration<T>(&self, label: &str, f: impl FnOnce() -> T) -> T {
-        self.stats.update_duration.observe_duration(label, f)
-    }
-
-    fn observe_size(&self, label: &str, rows: &[Row]) {
-        self.stats.update_size.observe(label, db_rows_size(rows));
-    }
-
-    fn report_stats(&self, batch: &WriteBatch) {
-        self.observe_size("write_funding_rows", &batch.funding_rows);
-        self.observe_size("write_spending_rows", &batch.spending_rows);
-        self.observe_size("write_txid_rows", &batch.txid_rows);
-        self.observe_size("write_header_rows", &batch.header_rows);
-        debug!(
-            "writing {} funding and {} spending rows from {} transactions, {} blocks",
-            batch.funding_rows.len(),
-            batch.spending_rows.len(),
-            batch.txid_rows.len(),
-            batch.header_rows.len()
-        );
-    }
-
-    pub(crate) fn sync(&mut self, daemon: &Daemon, exit_flag: &ExitFlag) -> Result<()> {
-        loop {
-            self.stats
-                .db_size
-                .set("total", self.store.get_size()? as f64);
-            let new_headers =
-                self.observe_duration("headers", || daemon.get_new_headers(&self.chain))?;
-            if new_headers.is_empty() {
-                break;
-            }
-            info!(
-                "indexing {} blocks: [{}..{}]",
-                new_headers.len(),
-                new_headers.first().unwrap().height(),
-                new_headers.last().unwrap().height()
-            );
-            for chunk in new_headers.chunks(self.batch_size) {
-                exit_flag.poll().with_context(|| {
-                    format!(
-                        "indexing interrupted at height: {}",
-                        chunk.first().unwrap().height()
-                    )
-                })?;
-                let blockhashes: Vec<BlockHash> = chunk.iter().map(|h| h.hash()).collect();
-                let mut heights = chunk.iter().map(|h| h.height());
-
-                let mut batch = WriteBatch::default();
-                daemon.for_blocks(blockhashes, |_blockhash, block| {
-                    let height = heights.next().expect("unexpected block");
-                    self.observe_duration("block", || {
-                        index_single_block(block, height).extend(&mut batch)
-                    });
-                    self.stats.height.set("tip", height as f64);
-                })?;
-                let heights: Vec<_> = heights.collect();
-                assert!(
-                    heights.is_empty(),
-                    "some blocks were not indexed: {:?}",
-                    heights
+    // Return `Ok(true)` when the chain is fully synced and the index is compacted.
+    pub(crate) fn sync(&mut self, daemon: &Daemon, exit_flag: &ExitFlag) -> Result<bool> {
+        let new_headers = self
+            .stats
+            .observe_duration("headers", || daemon.get_new_headers(&self.chain))?;
+        match (new_headers.first(), new_headers.last()) {
+            (Some(first), Some(last)) => {
+                let count = new_headers.len();
+                info!(
+                    "indexing {} blocks: [{}..{}]",
+                    count,
+                    first.height(),
+                    last.height()
                 );
-                batch.sort();
-                self.report_stats(&batch);
-                self.observe_duration("write", || self.store.write(batch));
-                self.stats
-                    .db_size
-                    .set("total", self.store.get_size()? as f64);
             }
-            self.chain.update(new_headers);
+            _ => {
+                self.store.flush(); // full compaction is performed on the first flush call
+                self.is_ready = true;
+                return Ok(true); // no more blocks to index (done for now)
+            }
         }
-        self.store.flush();
+        for chunk in new_headers.chunks(self.batch_size) {
+            exit_flag.poll().with_context(|| {
+                format!(
+                    "indexing interrupted at height: {}",
+                    chunk.first().unwrap().height()
+                )
+            })?;
+            self.sync_blocks(daemon, chunk)?;
+        }
+        self.chain.update(new_headers);
+        self.stats.observe_chain(&self.chain);
+        Ok(false) // sync is not done
+    }
+
+    fn sync_blocks(&mut self, daemon: &Daemon, chunk: &[NewHeader]) -> Result<()> {
+        let blockhashes: Vec<BlockHash> = chunk.iter().map(|h| h.hash()).collect();
+        let mut heights = chunk.iter().map(|h| h.height());
+
+        let mut batch = WriteBatch::default();
+        daemon.for_blocks(blockhashes, |_blockhash, block| {
+            let height = heights.next().expect("unexpected block");
+            self.stats.observe_duration("block", || {
+                index_single_block(block, height).extend(&mut batch)
+            });
+            self.stats.height.set("tip", height as f64);
+        })?;
+        let heights: Vec<_> = heights.collect();
+        assert!(
+            heights.is_empty(),
+            "some blocks were not indexed: {:?}",
+            heights
+        );
+        batch.sort();
+        self.stats.observe_batch(&batch);
+        self.stats
+            .observe_duration("write", || self.store.write(&batch));
+        self.stats.observe_db(&self.store);
         Ok(())
     }
+
+    pub(crate) fn is_ready(&self) -> bool {
+        self.is_ready
+    }
 }
 
 fn db_rows_size(rows: &[Row]) -> usize {
@@ -229,12 +252,12 @@ fn db_rows_size(rows: &[Row]) -> usize {
 }
 
 fn index_single_block(block: Block, height: usize) -> IndexResult {
-    let mut funding_rows = vec![];
-    let mut spending_rows = vec![];
+    let mut funding_rows = Vec::with_capacity(block.txdata.iter().map(|tx| tx.output.len()).sum());
+    let mut spending_rows = Vec::with_capacity(block.txdata.iter().map(|tx| tx.input.len()).sum());
     let mut txid_rows = Vec::with_capacity(block.txdata.len());
 
     for tx in &block.txdata {
-        txid_rows.push(TxidRow::new(tx.txid(), height));
+        txid_rows.push(TxidRow::row(tx.txid(), height));
 
         funding_rows.extend(
             tx.output
@@ -242,7 +265,7 @@ fn index_single_block(block: Block, height: usize) -> IndexResult {
                 .filter(|txo| !txo.script_pubkey.is_provably_unspendable())
                 .map(|txo| {
                     let scripthash = ScriptHash::new(&txo.script_pubkey);
-                    ScriptHashRow::new(scripthash, height)
+                    ScriptHashRow::row(scripthash, height)
                 }),
         );
 
@@ -252,7 +275,7 @@ fn index_single_block(block: Block, height: usize) -> IndexResult {
         spending_rows.extend(
             tx.input
                 .iter()
-                .map(|txin| SpendingPrefixRow::new(txin.previous_output, height)),
+                .map(|txin| SpendingPrefixRow::row(txin.previous_output, height)),
         );
     }
     IndexResult {
