@@ -22,6 +22,7 @@ use std::time::{SystemTime, UNIX_EPOCH};
 
 use crate::{
     chain::{Chain, NewHeader},
+    config::ELECTRS_VERSION,
     metrics::{default_duration_buckets, default_size_buckets, Histogram, Metrics},
 };
 
@@ -230,7 +231,7 @@ impl Connection {
                         Ok(msg) => msg,
                         Err(_) => {  // p2p_recv is closed, so rx_send is disconnected
                             debug!("closing p2p_loop thread: peer has disconnected");
-                            return Ok(());
+                            return Ok(()); // new_block_send is dropped, causing the server to exit
                         }
                     };
                     match msg {
@@ -243,7 +244,7 @@ impl Connection {
                         }
                         NetworkMessage::Inv(inventory) => {
                             debug!("peer inventory: {:?}", inventory);
-                            if inventory.iter().any(is_block_inv) {
+                            if inventory.iter().any(|inv| matches!(inv, Inventory::Block(_))) {
                                 let _ = new_block_send.try_send(()); // best-effort notification
                             }
 
@@ -257,6 +258,7 @@ impl Connection {
                         NetworkMessage::Block(block) => blocks_send.send(block)?,
                         NetworkMessage::Headers(headers) => headers_send.send(headers)?,
                         NetworkMessage::Alert(_) => (),  // https://bitcoin.org/en/alert/2016-11-01-alert-retirement
+                        NetworkMessage::Addr(_) => (),   // unused
                         msg => warn!("unexpected message: {:?}", msg),
                     }
                 }
@@ -305,20 +307,12 @@ fn build_version_message() -> NetworkMessage {
         receiver: address::Address::new(&addr, services),
         sender: address::Address::new(&addr, services),
         nonce: secp256k1::rand::thread_rng().gen(),
-        user_agent: String::from("electrs"),
+        user_agent: format!("/electrs:{}/", ELECTRS_VERSION),
         start_height: 0,
         relay: false,
     })
 }
 
-fn is_block_inv(inv: &Inventory) -> bool {
-    if let Inventory::Block(_) = inv {
-        true
-    } else {
-        false
-    }
-}
-
 struct RawNetworkMessage {
     magic: u32,
     cmd: CommandString,
@@ -346,6 +340,7 @@ impl RawNetworkMessage {
             "pong" => NetworkMessage::Pong(Decodable::consensus_decode(&mut raw)?),
             "reject" => NetworkMessage::Reject(Decodable::consensus_decode(&mut raw)?),
             "alert" => NetworkMessage::Alert(Decodable::consensus_decode(&mut raw)?),
+            "addr" => NetworkMessage::Addr(Decodable::consensus_decode(&mut raw)?),
             _ => NetworkMessage::Unknown {
                 command: self.cmd,
                 payload: self.raw,
