@@ -15,7 +15,6 @@ use crate::{
     daemon::Daemon,
     index::Index,
     mempool::Mempool,
-    merkle::Proof,
     types::{ScriptHash, StatusHash},
 };
 
@@ -42,7 +41,7 @@ impl TxEntry {
 
     /// Relevant (scripthash-wise) funded outpoints
     fn funding_outpoints(&self) -> impl Iterator<Item = OutPoint> + '_ {
-        make_outpoints(&self.txid, &self.outputs)
+        make_outpoints(self.txid, &self.outputs)
     }
 }
 
@@ -326,48 +325,48 @@ impl ScriptHashStatus {
         cache: &Cache,
         outpoints: &mut HashSet<OutPoint>,
     ) -> Result<HashMap<BlockHash, Vec<TxEntry>>> {
-        type TxPosition = usize; // transaction position within a block
-        let mut result = HashMap::<BlockHash, HashMap<TxPosition, TxEntry>>::new();
+        let scripthash = self.scripthash;
+        let mut result = HashMap::<BlockHash, HashMap<usize, TxEntry>>::new();
 
-        let funding_blockhashes = index.limit_result(index.filter_by_funding(self.scripthash))?;
+        let funding_blockhashes = index.limit_result(index.filter_by_funding(scripthash))?;
         self.for_new_blocks(funding_blockhashes, daemon, |blockhash, block| {
-            let txids: Vec<Txid> = block.txdata.iter().map(|tx| tx.txid()).collect();
-            for (pos, (tx, txid)) in block.txdata.into_iter().zip(txids.iter()).enumerate() {
-                let funding_outputs = filter_outputs(&tx, &self.scripthash);
-                if funding_outputs.is_empty() {
-                    continue;
-                }
-                cache.add_tx(*txid, move || tx);
-                cache.add_proof(blockhash, *txid, || Proof::create(&txids, pos));
-                outpoints.extend(make_outpoints(txid, &funding_outputs));
-                result
-                    .entry(blockhash)
-                    .or_default()
-                    .entry(pos)
-                    .or_insert_with(|| TxEntry::new(*txid))
-                    .outputs = funding_outputs;
-            }
+            let block_entries = result.entry(blockhash).or_default();
+            filter_block_txs(block, |tx| filter_outputs(&tx, scripthash)).for_each(
+                |FilteredTx {
+                     pos,
+                     tx,
+                     txid,
+                     result: funding_outputs,
+                 }| {
+                    cache.add_tx(txid, move || tx);
+                    outpoints.extend(make_outpoints(txid, &funding_outputs));
+                    block_entries
+                        .entry(pos)
+                        .or_insert_with(|| TxEntry::new(txid))
+                        .outputs = funding_outputs;
+                },
+            );
         })?;
         let spending_blockhashes: HashSet<BlockHash> = outpoints
             .par_iter()
             .flat_map_iter(|outpoint| index.filter_by_spending(*outpoint))
             .collect();
         self.for_new_blocks(spending_blockhashes, daemon, |blockhash, block| {
-            let txids: Vec<Txid> = block.txdata.iter().map(|tx| tx.txid()).collect();
-            for (pos, (tx, txid)) in block.txdata.into_iter().zip(txids.iter()).enumerate() {
-                let spent_outpoints = filter_inputs(&tx, outpoints);
-                if spent_outpoints.is_empty() {
-                    continue;
-                }
-                cache.add_tx(*txid, move || tx);
-                cache.add_proof(blockhash, *txid, || Proof::create(&txids, pos));
-                result
-                    .entry(blockhash)
-                    .or_default()
-                    .entry(pos)
-                    .or_insert_with(|| TxEntry::new(*txid))
-                    .spent = spent_outpoints;
-            }
+            let block_entries = result.entry(blockhash).or_default();
+            filter_block_txs(block, |tx| filter_inputs(&tx, outpoints)).for_each(
+                |FilteredTx {
+                     pos,
+                     tx,
+                     txid,
+                     result: spent_outpoints,
+                 }| {
+                    cache.add_tx(txid, move || tx);
+                    block_entries
+                        .entry(pos)
+                        .or_insert_with(|| TxEntry::new(txid))
+                        .spent = spent_outpoints;
+                },
+            );
         })?;
 
         Ok(result
@@ -376,7 +375,7 @@ impl ScriptHashStatus {
                 // sort transactions by their position in a block
                 let sorted_entries = entries_map
                     .into_iter()
-                    .collect::<BTreeMap<TxPosition, TxEntry>>()
+                    .collect::<BTreeMap<usize, TxEntry>>()
                     .into_iter()
                     .map(|(_pos, entry)| entry)
                     .collect::<Vec<TxEntry>>();
@@ -395,9 +394,9 @@ impl ScriptHashStatus {
     ) -> Vec<TxEntry> {
         let mut result = HashMap::<Txid, TxEntry>::new();
         for entry in mempool.filter_by_funding(&self.scripthash) {
-            let funding_outputs = filter_outputs(&entry.tx, &self.scripthash);
+            let funding_outputs = filter_outputs(&entry.tx, self.scripthash);
             assert!(!funding_outputs.is_empty());
-            outpoints.extend(make_outpoints(&entry.txid, &funding_outputs));
+            outpoints.extend(make_outpoints(entry.txid, &funding_outputs));
             result
                 .entry(entry.txid)
                 .or_insert_with(|| TxEntry::new(entry.txid))
@@ -462,20 +461,17 @@ impl ScriptHashStatus {
     }
 }
 
-fn make_outpoints<'a>(
-    txid: &'a Txid,
-    outputs: &'a [TxOutput],
-) -> impl Iterator<Item = OutPoint> + 'a {
+fn make_outpoints<'a>(txid: Txid, outputs: &'a [TxOutput]) -> impl Iterator<Item = OutPoint> + 'a {
     outputs
         .iter()
-        .map(move |out| OutPoint::new(*txid, out.index))
+        .map(move |out| OutPoint::new(txid, out.index))
 }
 
-fn filter_outputs(tx: &Transaction, scripthash: &ScriptHash) -> Vec<TxOutput> {
+fn filter_outputs(tx: &Transaction, scripthash: ScriptHash) -> Vec<TxOutput> {
     let outputs = tx.output.iter().zip(0u32..);
     outputs
         .filter_map(move |(txo, vout)| {
-            if ScriptHash::new(&txo.script_pubkey) == *scripthash {
+            if ScriptHash::new(&txo.script_pubkey) == scripthash {
                 Some(TxOutput {
                     index: vout,
                     value: Amount::from_sat(txo.value),
@@ -511,6 +507,38 @@ fn compute_status_hash(history: &[HistoryEntry]) -> Option<StatusHash> {
     Some(StatusHash::from_engine(engine))
 }
 
+struct FilteredTx<T> {
+    tx: Transaction,
+    txid: Txid,
+    pos: usize,
+    result: Vec<T>,
+}
+
+fn filter_block_txs<T: Send>(
+    block: Block,
+    map_fn: impl Fn(&Transaction) -> Vec<T> + Sync,
+) -> impl Iterator<Item = FilteredTx<T>> {
+    block
+        .txdata
+        .into_par_iter()
+        .enumerate()
+        .filter_map(|(pos, tx)| {
+            let result = map_fn(&tx);
+            if result.is_empty() {
+                return None; // skip irrelevant transaction
+            }
+            let txid = tx.txid();
+            Some(FilteredTx {
+                tx,
+                txid,
+                pos,
+                result,
+            })
+        })
+        .collect::<Vec<_>>()
+        .into_iter()
+}
+
 #[cfg(test)]
 mod tests {
     use super::HistoryEntry;
